import logging
from datetime import datetime
from typing import Optional

from pandas import DataFrame

from monopoly.config import (
    BruteForceConfig,
    PdfConfig,
    StatementConfig,
    TransactionConfig,
)
from monopoly.constants import StatementFields
from monopoly.csv import write_to_csv
from monopoly.pdf import PdfParser
from monopoly.statement import Statement

logger = logging.getLogger(__name__)


class StatementProcessor(PdfParser):
    statement_config: StatementConfig
    transaction_config: TransactionConfig
    pdf_config: Optional[PdfConfig] = None
    brute_force_config: Optional[BruteForceConfig] = None
    parser: Optional[PdfParser] = None
    safety_check_enabled: bool = True

    """
    Handles extract, transform and load (ETL) logic for bank statements

    Transactions are extracted from pages, before undergoing various
    transformations like converting various date formats into ISO 8601.
    Transformed statements are then written to a CSV file.

    Since the auto-detect function already creates a parser, this class
    allows for the parser to be reused and avoid re-opening the PDF.
    """

    def __init__(self, file_path: str, parser: Optional[PdfParser] = None, **kwargs):
        keys = [
            "statement_config",
            "transaction_config",
            "pdf_config",
            "brute_force_config",
            "safety_check_enabled",
        ]

        for key, value in kwargs.items():
            if key in keys:
                setattr(self, key, value)

        if not parser:
            super().__init__(file_path, self.brute_force_config, self.pdf_config)

    def extract(self) -> Statement:
        """Extracts transactions from the statement, and performs
        a safety check to make sure that total transactions add up"""
        pages = self.get_pages(self.brute_force_config)
        statement = Statement(pages, self.statement_config, self.transaction_config)

        if not statement.transactions:
            raise ValueError("No transactions found - statement extraction failed")

        if not statement.statement_date:
            raise ValueError("No statement date found")

        if self.safety_check_enabled:
            result = self._perform_safety_check(statement)
            if not result:
                logger.warning(
                    "Unable to verify that all transactions have been extracted"
                )

        return statement

    def _perform_safety_check(self, statement: Statement) -> bool:
        """Checks that the total sum of all transactions is present
        somewhere within the document

        Text is re-extracted from the page, as some bank-specific bounding-box
        configurations (e.g. HSBC) may preclude the total from being extracted

        Returns `False` if the total does not exist in the document.
        """
        decimal_numbers = set()
        for page in self.document:
            lines = page.get_textpage().extractText().split("\n")
            decimal_numbers.update(statement.get_decimal_numbers(lines))
        total_amount = round(statement.df[StatementFields.AMOUNT].sum(), 2)
        return total_amount in decimal_numbers

    def transform(self, statement: Statement) -> DataFrame:
        logger.debug("Running transformation functions on DataFrame")

        df = statement.df
        statement_date = statement.statement_date
        date_format = self.transaction_config.date_format

        def convert_date(row):
            parsed_date = datetime.strptime(
                row[StatementFields.TRANSACTION_DATE], date_format
            )
            row_day, row_month = parsed_date.day, parsed_date.month

            if statement_date.month == 1 and row_month == 12:
                row_year = statement_date.year - 1
            else:
                row_year = statement_date.year

            return f"{row_year}-{row_month:02d}-{row_day:02d}"

        logger.debug("Transforming dates to ISO 8601")
        df[StatementFields.TRANSACTION_DATE] = df.apply(convert_date, axis=1)
        return df

    def load(
        self, df: DataFrame, statement: Statement, csv_file_path: Optional[str] = None
    ):
        csv_file_path = write_to_csv(
            df=df, csv_file_path=csv_file_path, statement=statement
        )

        return csv_file_path



from typing import Annotated, Optional

from pydantic import StringConstraints
from pydantic.dataclasses import dataclass
from pydantic_settings import BaseSettings, SettingsConfigDict

from monopoly.constants import AccountType, BankNames


class Settings(BaseSettings):
    """
    Pydantic model that automatically populates variables from a .env file.

    HSBC passwords differ per card (since HSBC uses the last 6 digits of each card).
    If only a single statement from HSBC needs to be parsed, the hsbc_pdf_password
    can be used.

    If there are multiple statements from HSBC, the `hsbc_pdf_password_prefix` can
    be used to automatically unlock PDFs, which is used as a static string
    in the HSBC `brute_force_config`.
    """

    ocbc_pdf_password: str = ""
    citibank_pdf_password: str = ""
    standard_chartered_pdf_password: str = ""
    hsbc_pdf_password: str = ""
    hsbc_pdf_password_prefix: str = ""

    model_config = SettingsConfigDict(env_file=".env", extra="allow")


@dataclass
class TransactionConfig:
    """
    Stores configurations for transactions within the statement

    - `pattern` refers to the regex pattern used to capture transactions,
    where a pattern like:
        "(?P<transaction_date>\\d+/\\d+)\\s*"
        "(?P<description>.*?)\\s*"
        "(?P<amount>[\\d.,]+)$"
    is used to capture a transaction like:
        06/07 URBAN TRANSIT CO. SINGAPORE SG  1.38
    - `date_format` represents the datetime format that a specific bank uses
    for transactions. For example, "%d/%m" will match 06/07
    - `cashback_key` is the identifier that is used to differentiate cashback
    transactions from items we don't want to capture like AXS bill payments.
    e.g. if the cashback_key="CASH REBATE", we'll treat transactions containing
    "CASH REBATE" as cashback transactions, and include it in the statement
    - `multiline_transactions` controls whether Monopoly tries to concatenate
    transactions that are split across two lines
    """

    pattern: str
    date_format: Annotated[str, StringConstraints(pattern="%")]
    cashback_key: Optional[str] = None
    multiline_transactions: bool = False


@dataclass
class StatementConfig:
    """
    Stores configuration values for the `Statement` class
    """

    bank_name: BankNames
    account_type: AccountType
    statement_date_format: Annotated[str, StringConstraints(pattern="%.+%.+%")]
    statement_date_pattern: str

    # Convert enums to strings
    def __post_init__(self):
        self.bank_name = self.bank_name.value
        self.account_type = self.account_type.value


@dataclass
class PdfConfig:
    """Stores PDF configuration values for the `PdfParser` class

    - `password`: The password used to unlock the PDF (if it is locked)
    - `page_range`: A slice representing which pages to process. For
    example, a range of (1, -1) will mean that the first and last pages
    are skipped.
    - `page_bbox`: A tuple representing the bounding box range for every
    page. This is used to avoid weirdness like vertical text, and other
    PDF artifacts that may affect parsing.
    """

    password: Optional[str] = None
    page_range: tuple[Optional[int], Optional[int]] = (None, None)
    page_bbox: Optional[tuple[float, float, float, float]] = None


@dataclass
class BruteForceConfig:
    """
    Stores configuration values to help automatically unlock PDFs.

    Attributes:
    - `static_string`: A static string used in PDF passwords.

      Example: For a password format like "01Jan1999123456," the static string
      might be "01Jan1999," which remains constant across all HSBC statements
      for a particular user.

    - `mask`: A mask used for generating variations in the password.

      Example: In the same password format "01Jan1999123456," the mask
      represents the last six digits (e.g., "123456"). The masking pattern
      used in this case should be `?d?d?d?d?d?d`.

    These attributes are used together to automatically unlock PDFs where a portion
    of the password remains consistent (the static string) while another part varies
    (the brute force mask).
    """

    static_string: Optional[str] = None
    mask: Optional[str] = None


settings = Settings()



from monopoly.banks.base import BankBase
from monopoly.config import StatementConfig, TransactionConfig
from monopoly.constants import AccountType, BankNames


class MonopolyBank(BankBase):
    """Dummy class to help with reading the example PDF statement"""

    statement_config = StatementConfig(
        bank_name=BankNames.EXAMPLE,
        account_type=AccountType.CREDIT,
        statement_date_pattern=r"\d{2}\-\d{2}\-\d{4}",
        statement_date_format=r"%d-%m-%Y",
    )

    transaction_config = TransactionConfig(
        pattern=(
            r"(?P<transaction_date>\d+/\d+)\s*"
            r"(?P<description>.*?)\s*"
            r"(?P<amount>[\d.,]+)$"
        ),
        date_format=r"%d/%m",
    )



import logging
import subprocess
from dataclasses import dataclass
from functools import cached_property
from io import BytesIO
from typing import Optional

import fitz
import pdftotext
from pdf2john import PdfHashExtractor

from monopoly.config import BruteForceConfig, PdfConfig

logger = logging.getLogger(__name__)


@dataclass
class PdfPage:
    """
    Dataclass representation of a bank statement PDF page.
    Contains the raw text of a PDF page, and allows access
    to the raw text as a list via the `lines` property
    """

    raw_text: str

    @cached_property
    def lines(self) -> list[str]:
        return list(filter(None, self.raw_text.split("\n")))


class PdfParser:
    def __init__(
        self,
        file_path: str,
        brute_force_config: Optional[BruteForceConfig] = None,
        pdf_config: Optional[PdfConfig] = None,
    ):
        """
        Class responsible for parsing PDFs and returning raw text

        The page_range variable determines which pages are extracted.
        All pages are extracted by default.
        """
        self.file_path = file_path

        if pdf_config is None:
            pdf_config = PdfConfig()

        self.password = pdf_config.password
        self.page_range = slice(*pdf_config.page_range)
        self.page_bbox = pdf_config.page_bbox
        self.brute_force_config = brute_force_config

    def open(self, brute_force_config: Optional[BruteForceConfig] = None):
        """
        Opens and decrypts a PDF document
        """
        logger.info("Opening pdf from path %s", self.file_path)
        document = self.document

        if not document.is_encrypted:
            return document

        if self.password:
            document.authenticate(self.password)

            if document.is_encrypted:
                raise ValueError("Wrong password - unable to open document")

            return document

        # This attempts to unlock statements based on a common password,
        # followed by the last few digits of a card
        if brute_force_config:
            logger.info("Unlocking PDF using a string prefix with mask")
            password = self.unlock_pdf(
                static_string=brute_force_config.static_string,
                mask=brute_force_config.mask,
            )

            document.authenticate(password)

            if not document.is_encrypted:
                logger.debug("Successfully authenticated with password")
                return document

            # If no successful authentication, raise an error
            raise ValueError(
                "Unable to unlock PDF password using static string and mask"
            )

        if document.is_encrypted:
            raise RuntimeError("Document failed to open -- check password")

        raise RuntimeError("Failed to open document")

    def get_pages(self, brute_force_config=None) -> list[PdfPage]:
        logger.debug("Extracting text from PDF")
        document: fitz.Document = self.open(brute_force_config)

        num_pages = list(range(document.page_count))
        document.select(num_pages[self.page_range])

        for page in document:
            if self.page_bbox:
                logger.debug("Cropping page")
                page.set_cropbox(self.page_bbox)

            logger.debug("Removing vertical text")
            page = self._remove_vertical_text(page)

        pdf_byte_stream = BytesIO(document.tobytes())
        pdf = pdftotext.PDF(pdf_byte_stream, physical=True)

        return [PdfPage(page) for page in pdf]

    @cached_property
    def document(self) -> fitz.Document:
        """
        Returns a Python representation of a PDF document.
        """
        return fitz.Document(self.file_path)

    @cached_property
    def extractor(self) -> PdfHashExtractor:
        """
        Returns an instance of a PDF hash extractor, used
        to read encryption metadata and generate a password hash
        """
        return PdfHashExtractor(self.file_path)

    @staticmethod
    def _remove_vertical_text(page: fitz.Page):
        """Helper function to remove vertical text, based on writing direction (wdir).

        This helps avoid situations where the PDF is oddly parsed, due to vertical text
        inside the PDF.

        An example of vertical text breaking a transaction:
        ```
        'HEALTHY HARVEST CAFÉ SINGAPORE SG',
        'Co Reg No: 123456',
        '10 NOV 9.80',
        ```

        Note:
            The 'dir' key represents the tuple (cosine, sine) for the angle.
            If line["dir"] != (1, 0), the text of its spans is rotated.

        """
        for block in page.get_text("dict", flags=fitz.TEXTFLAGS_TEXT)["blocks"]:
            for line in block["lines"]:
                writing_direction = line["dir"]
                if writing_direction != (1, 0):
                    page.add_redact_annot(line["bbox"])
        page.apply_redactions(images=fitz.PDF_REDACT_IMAGE_NONE)
        return page

    def unlock_pdf(self, static_string: Optional[str], mask: Optional[str]):
        """
        Extracts the hashed representation for a set of PDF passwords (owner/user),
        and attempts to automatically unlock/decrypt the PDF based on
        the static string and mask using `john`
        """
        hash_extractor = self.extractor
        pdf_hash = hash_extractor.parse()

        hash_path = ".hash"
        with open(hash_path, "w", encoding="utf-8") as file:
            file.write(pdf_hash)

        mask_command = [
            f"john --format=PDF --mask={static_string}{mask} {hash_path} --pot=.pot"
        ]
        process = subprocess.run(
            mask_command,
            shell=True,
            check=False,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )

        if not process.returncode == 0:
            raise ValueError(f"Return code is not 0: {process}")

        show_command = ["john", "--show", hash_path, "--pot=.pot"]
        output = subprocess.run(
            show_command,
            text=True,
            check=False,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )

        if not output.returncode == 0:
            raise ValueError(f"Return code is not 0: {output}")

        if "1 password hash cracked, 0 left" not in output.stdout:
            raise ValueError(f"PDF was not unlocked: {output}")

        password = output.stdout.split("\n")[0].split(":")[-1]

        return password



import logging
import re
from datetime import datetime
from functools import cached_property

from pandas import DataFrame
from pydantic import field_validator
from pydantic.dataclasses import dataclass

from monopoly.config import StatementConfig, TransactionConfig
from monopoly.constants import StatementFields
from monopoly.pdf import PdfPage

logger = logging.getLogger(__name__)


@dataclass
class Transaction:
    """
    Holds transaction data, validates the data, and
    performs various coercions like removing whitespace

    The class attributes are consistent with the enum
    values from StatementFields
    """

    transaction_date: str
    description: str
    amount: float

    @field_validator("description", mode="after")
    def remove_extra_whitespace(cls, value: str) -> str:
        return " ".join(value.split())

    @field_validator("amount", mode="before")
    def adjust_amount(cls, value: str) -> str:
        """
        Replaces commas, whitespaces and parentheses in string representation of floats
        e.g.
            1,234.00 -> 1234.00
            (-10.00) -> -10.00
            (-1.56 ) -> -1.56
        """
        if isinstance(value, str):
            # change cashback to negative transaction
            if value.startswith("(") and value.endswith(")"):
                value = "-" + value
            return re.sub(r"[,)(\s]", "", value)
        return value


@dataclass
class Statement:
    """
    A dataclass representation of a bank statement, containing
    the PDF pages (their raw text representation in a `list`),
    and specific config per bank.
    """

    pages: list[PdfPage]
    columns = [enum.value for enum in StatementFields]
    statement_config: StatementConfig
    transaction_config: TransactionConfig

    @cached_property
    def transactions(self) -> list[Transaction]:
        transactions = []
        for page in self.pages:
            lines = self.process_lines(page)
            for i, line in enumerate(lines):
                transaction = self._process_line(
                    self.transaction_config, line, lines, idx=i
                )
                if transaction:
                    transactions.append(transaction)

        return transactions

    @staticmethod
    def process_lines(page: PdfPage) -> list:
        return [line.lstrip() for line in page.lines]

    @staticmethod
    def _process_line(
        transaction_config: TransactionConfig, line: str, lines: list[str], idx: int
    ) -> Transaction | None:
        if match := re.search(transaction_config.pattern, line):
            # if the cashback key isn't in the description, it's not
            # a cashback transaction, and is of no interest to us
            if Statement.is_enclosed_in_parentheses(match["amount"]):
                if transaction_config.cashback_key not in match["description"]:
                    return None

            transaction = Transaction(**match.groupdict())  # type: ignore

            # handle transactions that span multiple lines
            if transaction_config.multiline_transactions and idx < len(lines) - 1:
                next_line = lines[idx + 1]
                if not re.search(transaction_config.pattern, next_line):
                    transaction.description += " " + next_line
                    transaction = Transaction(**vars(transaction))
            return transaction
        return None

    @staticmethod
    def is_enclosed_in_parentheses(amount: str) -> bool:
        return amount.startswith("(") and amount.endswith(")")

    @cached_property
    def statement_date(self):
        config = self.statement_config
        logger.debug("Extracting statement date")
        first_page = self.pages[0]
        for line in first_page.lines:
            if match := re.findall(config.statement_date_pattern, line):
                statement_date = match[0]
                logger.debug("Statement date found")
                return datetime.strptime(statement_date, config.statement_date_format)
        return None

    @staticmethod
    def get_decimal_numbers(lines: list[str]) -> set[float]:
        """Returns all decimal numbers in a statement. This is used
        to perform a safety check, to make sure no transactions have been missed"""

        number_pattern = re.compile(r"[\d.,]+")
        decimal_pattern = re.compile(r"\d+\.\d+$")
        numbers = set()
        for line in lines:
            numbers.update(re.findall(number_pattern, line))
            numbers = {number.replace(",", "") for number in numbers}
            decimal_numbers = {
                float(number) for number in numbers if decimal_pattern.match(number)
            }
        return decimal_numbers

    @cached_property
    def df(self) -> DataFrame:
        return DataFrame(self.transactions, columns=self.columns)



from dataclasses import Field, fields
from typing import Type

from monopoly.constants import EncryptionIdentifier, MetadataIdentifier
from monopoly.pdf import PdfParser

from .base import BankBase
from .citibank import Citibank
from .dbs import Dbs
from .hsbc import Hsbc
from .ocbc import Ocbc
from .standard_chartered import StandardChartered

banks: list[Type[BankBase]] = [Citibank, Dbs, Hsbc, Ocbc, StandardChartered]


def auto_detect_bank(file_path: str) -> BankBase:
    """
    Reads the encryption metadata or actual metadata (if the PDF is not encrypted),
    and checks for a bank based on unique identifiers.
    """
    parser = PdfParser(file_path)
    for bank in banks:
        metadata = bank.get_identifier(parser)

        if is_bank_identified(metadata, bank):
            return bank(file_path=parser.file_path, parser=parser)

    raise ValueError(
        f"Could not find a bank for {parser.file_path}."
        f"This may be due a bank PDF update, leading to obsolete identifiers."
        f"To get around this, create a manual instance of the bank class."
    )


def is_bank_identified(metadata, bank: Type[BankBase]) -> bool:
    if bank.identifiers:
        for identifier in bank.identifiers:
            if all(
                check_identifier_field(field, metadata, identifier)
                for field in fields(metadata)
            ):
                return True

    return False


def check_identifier_field(
    field: Field,
    metadata: EncryptionIdentifier | MetadataIdentifier,
    identifier: EncryptionIdentifier | MetadataIdentifier,
):
    # Only compare matching identifier types
    if type(metadata) is type(identifier):
        field_value = getattr(metadata, field.name)
        identifier_value = getattr(identifier, field.name)

        # allow for partial string matching
        if isinstance(field.type(), str):
            return identifier_value in field_value

        # otherwise check that values match exactly
        return identifier_value == field_value
    return False
